<ul>{% for calendar in calendar_list %}
  <li>Server "{{calendar.server}}"
    <ul>
      <li>CALENDAR = {{calendar.name}}</li>
      <li>ID = {{calendar.id}}</li>
      <li>COMPONENTS = {{', '.join(calendar.components)}}</li>
      <li>URL = {{calendar.url}}</li>
    </ul>
  </li>
{% endfor %}</ul>
